<CsoundSynthesizer>

Hammond Novachord Emulation
26 Feb 2019 -- with non-Novachord Overdrive and Reverb...
28 Feb 2019 -- Optional Velocity Sensitivity and Aftertouch
4 Mar 2019 -- Macros for dB range, aftertouch, and detuning amount
8 Mar Note changed CCs!
5 Apr 2019 -- MIDI RPN tuning adjustment and prevention of overlaid notes
<CsOptions>
-m0 -d -odac -Lstdin -M0
</CsOptions>

<CsInstruments>

; Initialize the global variables.
sr = 48000
;kr = 960
ksmps = 120
nchnls = 2

0dbfs = 1

/************************************************************
 MIDI Assignments:
*************************************************************/

giChan = 1

massign giChan,"Generator"	; Keyboard and Controls
massign 10, "Params"	; Percussion event triggers parameter saving
pgmassign 0, 0	; Ignore Program Changes

giExprCC = 11	; Expression Pedal

giHoldCC = 66	; "Sustenuto" controller used for Novachord 'Sustain' mode
giHoldModeCC = 5	; (Fader 7) only use above if this CC has value zero

; Tone Controls (Axiom Patch 1 conntrols in parentheses)
giDeepCC = 74	;Deep Tone (Fader 1)
giRes1CC = 71	;Resonator 1 (Fader 2)
giRes2CC = 91	;Resonator 2 (Fader 3)
giRes3CC = 93	;Resonator 3 (Fader 4)
giBrillCC = 73	;Brilliant Tone (Fader 5)
giFullCC = 72	;Full Tone (Fader 6)
giEnvlpCC = 84	;Select Attack Mode (7 positions + 0) (Fader 8)

;Knob Controls (Axiom) -- lower row: (Patch 1)
giAtkCC = 10	; Attack Time -- 0.2..0.01 secs [Note reversed direction!]
giSusCC = 77	; Sustain Level -- 1.0..0.1 of peak [  "        "        ]
giRelCC = 78	; Release Time -- 0.1..2.0 secs (after key release)
giDcyCC = 79	; Decay Time -- 0.01..0.5 secs (to Sustain level after Attack)

giVibCC = 1		;Mod Wheel for Vibrato

giRvbCC = 75	; Reverb (Knob 1 -- upper row)
giOvrdrvCC = 76	;Overdrive Distortion (knob 2)

/************************************************************
 Compute velocity curve and set default macros:
 -- change these from the command line if desired
*************************************************************/

#ifndef DBRANGE
 #define DBRANGE #60#
#end

#ifndef VOLMAX
 #define VOLMAX #1.0#
#end

iVRdB = $DBRANGE	; Range in dB of output for velocity range 1..127

iVR = 10^(iVRdB/20)
giVb = 127/(126*iVR^0.5)-1.0/126
giVm = (1-giVb)/127
printf_i "for dB range %g: iVb = %g iVm = %g\n", 1, iVRdB, giVb, giVm
;use amp = (giVm*vel + giVb)^2

;Vibrato Aftertouch amount (10 probably good if you want it)
#ifndef VTOUCH
 #define VTOUCH #0#
#end
;Amplitude  Aftertouch amount (max 1)
#ifndef ATOUCH
 #define ATOUCH #0#
#end

gktouchused init 0	; Fix Csound aftertouch bug (full-on if never used!)

/**************************************************************
Controller Initialization:
**************************************************************/

;  0..127 =>    0..1        0..1          0..1           0..1           0..1         0..1
ctrlinit 1, giFullCC, 64, giRes1CC, 120, giRes2CC, 120, giRes3CC, 32, giDeepCC, 0, giBrillCC, 0
;  0..127 => 1.0..0.1    0.2..0.01   0.1..2.0     0.01..0.5     0..7
ctrlinit 1, giSusCC, 0, giAtkCC, 0, giRelCC, 60, giDcyCC, 64, giEnvlpCC, 0
;  0..127 =>   0.0..1.0     0.0..20.0   0..0.8      0..4           Off/On
ctrlinit 1, giExprCC, 120, giVibCC, 0, giRvbCC, 0, giOvrdrvCC, 0, giHoldModeCC, 127

/*************************************************************/


giSqr ftgen 3, 0, 16384, 10,  1, 0, 0.3333, 0, 0.2, 0, 0.14285, 0, 0.11111, 0, 0.09091	;square
; Actual Waveform attempt (from patent 2,126,662 p.5 para.55):
giNovaWF ftgen 5, 0, 8193, 16, -1.0, 6144, 0, -1.0, 1536, 0, 1.0, 512, 0, -1.0

;; Expression Response Curve (from Hammond...):
giExpr ftgen 11, 0, 129, 5, 0.0001, 128, 1.0

;; Track multiple keys
giKeys ftgen 0, 0, 128, 2, 0


/**************************************************************
Oscillator Detuning:
-- by default <10 cents
**************************************************************/
giDetune ftgen  22, 0, 16, -02, 0.99612, 0.99878, 1.00368, 1.00231, 0.99561, 1.00294, 0.99672, 1.00281, 0.99466, 1.00488, 1.00166, 0.99895
;giDetune ftgen  22, 0, 16, -07, 1.0, 16, 1.0	; in tune

#ifdef DETUNE
indxdt = 0
icents = $DETUNE
	prints "Detuning up to %f cents\n", icents
	prints "Index \t cents   \t  frac\n"
loop:
	ir random -icents, icents
	idetune = 2^(ir/1200.0)
	tableiw idetune, indxdt, giDetune
	prints "%d:\t%f\t%f\n", indxdt, ir, idetune
	loop_lt indxdt, 1, 12, loop
#end

/*************************************************************/

giRvbRamp ftgen 0, 0, 129, 7, 0, 20, 0.5, 109, 1	; Room reverb is negligible below 0.6

gkvib0 init 0
gkvib1 init 0
gkvib2 init 0
gkvib3 init 0
gkvib4 init 0
gkvib5 init 0

gkVamp init 0.0
gkVtouch init 0.0

gkTuning init 1.0

gaout init 0

alwayson "Tuning"
alwayson "Vibrato"
alwayson "Control"
alwayson "Output"


instr Generator
	midinoteonkey, p4, p5	; key, vel
;	iamp = p5 *0dbfs/127	; for linear velocity (Real Novachord has no sensitivity)
	iamp = $VOLMAX*(giVm*p5 + giVb)^2	; for square-law velocity
	icps = cpsmidinn(p4)
	key init p4
	
;	printf "key = %d, freq = %g\n", 1, p4, icps

	; Cancel any earlier presses of this key:
	ikdn table p4, giKeys		;at init
	kmark init ikdn+1
	kinit init 1
	if kinit > 0 then
		tablew kmark, key, giKeys
		kinit = 0
	endif
	kdn table key, giKeys		;now
	kstfu init 0
	if (kdn > kmark) || (kstfu > 0)  then
 		kstfu = 1
		;turnoff
	else
		krel release
		if krel > 0 then
			tablew 0, key, giKeys	; writes at krate!
		endif

		iT times

		iActive active 1
		if (iActive) == 1 then
			giBaseTime = int(iT)
		endif

		iT = iT - giBaseTime	; keep ref time short for computation accuracy

;		prints "entry with ps: %f %f %f %f %f %f %f\n", p1, p2, p3, p4, p5, p6, p7
;		prints "instr 1 amp %f freq %f from %d %d\n", iamp, icps, p5, p4

		iatk midic7 giAtkCC, 0.2, 0.01 
;		printf " Attack = %f secs\n", iatk, iatk

		idcy midic7 giDcyCC, 0.01,1.0
;		printf " Decay = %f secs\n", idcy, idcy

		isus midic7 giSusCC, 1.0, 0.1
;		printf " Sustain level = %f\n", isus, isus

		irel midic7 giRelCC, 0.1, 2.0
;		printf " Release = %f secs\n", irel, irel
;		printf_i " Release(2) = %f secs\n", 1, irel

		ienv midic7 giEnvlpCC, 0, 7
		ienv = int(ienv)
;		printf_i " Envelope = %d\n", 1, ienv

		imode midic7 giHoldModeCC, 0, 2	;no cutoff on key-up if non-zero
		koff release
		khold init 0
		if koff == 0 then ; only if active before release
			khold midic7 giHoldCC, 0, 1
		endif
		khold = round(khold) + round(imode)

		kTouch aftouch 0, 1	
;		printf "aftertouch %f\n", kTouch, kTouch

		;;;;;;;;; Fix for Aftertouch bug -- must actually become zero before it's valid
		if kTouch < 0.5 then
			gktouchused = 1
		endif
		if gktouchused == 0 then
			kTouch = 0
		endif
		;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
	
;		gkVtouch = kTouch*$VTOUCH
		gkVtouch = kTouch^2*$VTOUCH	; square law probably best
		katouch tonek kTouch^2*$ATOUCH, 5	; square law probably best -- but smoothed

		idecaytim = 0.5
		ireleasetim = 1.5

		; These sort of match the published 'Attack" control ADSRs
		if ienv == 1 then
			kamp linsegr 1, idecaytim, 0, 0, 0
		elseif ienv == 2 then
			kamp linsegr 1, idecaytim, 0.2, ireleasetim, 0
		elseif ienv == 3 then
			kamp linsegr 0.7, idecaytim, 0.4, ireleasetim/3, 0
		elseif ienv == 4 then
			kamp linsegr 0.5, idecaytim, 0.5, ireleasetim/2, 0
		elseif ienv == 5 then
			kamp linsegr 0.3, idecaytim, 0.7, ireleasetim, 0
		elseif ienv == 6 then
			kamp linsegr 0.1, idecaytim, 0.8, ireleasetim, 0
		elseif ienv == 7 then
			kamp linsegr 0.0, idecaytim, 1.0, ireleasetim, 0
		else	; use completely adjustable parameters 
			kamp madsr iatk, idcy, isus, irel
		endif

		if (khold == 0 || kamp == 0) && koff == 1 then
;			printf " Silencing %d\n", koff, koff
			kamp = 0
		endif

		iKey = int(p4)%12
		idetune table iKey, giDetune
;		prints "key=%f scaling=%f\n", iKey, idetune

		kamp = kamp*(1+katouch)*iamp
		ivibno = int(p4/2)%6
;		prints "ivibno = %f for key = %d\n", ivibno, p4
		kvib = ivibno == 0 ? gkvib0 : ivibno == 1 ? gkvib1 :\
		 ivibno == 2 ? gkvib2 : ivibno == 3 ? gkvib3 :\
		 ivibno == 4 ? gkvib4 : gkvib5
;		kcps = icps + kvib*icps/1200
		kcps = icps*idetune + kvib*icps/1200
;		araw oscili kamp, kcps, iWave, frac(iT*icps)
		araw oscili kamp, kcps*gkTuning, giNovaWF, frac(iT*icps)
		araw dcblock2 araw
		gaout = gaout + araw
	endif	; of mult-key protection
endin


instr Tuning
	kstat, kchn, kd1, kd2 midiin
	if kstat == 176 then
		if kd1 == 101 then
			kcc101 = kd2
		endif
		if kd1 == 100 then
			kcc100 = kd2
		endif
		if kd1 == 6 && kcc101 == 0 && kcc100 == 1 then
			ktcents = (kd2 - 64)*100/64
			gkTuning = 2^(ktcents/1200.0)
		endif
	endif
endin

instr Vibrato
	/* actual vibrato is with reeds switching condensers: */
	kVamp = gkVamp + gkVtouch
	gkvib0 oscil kVamp, 8.7, giSqr
	gkvib1 oscil kVamp, 7.1, giSqr
	gkvib2 oscil kVamp, 7.4, giSqr
	gkvib3 oscil kVamp, 7.7, giSqr
	gkvib4 oscil kVamp, 5.0, giSqr
	gkvib5 oscil kVamp, 5.2, giSqr
endin


instr Control
	kdeepamp ctrl7 giChan, giDeepCC, 0.0, 1.0;, giExpr
	kdeepamp tonek kdeepamp, 10 
;	printf " Deep amp=%f\n", kdeepamp, kdeepamp

	kres1amp ctrl7 giChan, giRes1CC, 0.0, 1.0;, giExpr
	kres1amp tonek kres1amp, 10 
;	printf " Resonator 1 amp=%f\n", kres1amp, kres1amp

	kres2amp ctrl7 giChan, giRes2CC, 0.0, 1.0;, giExpr
	kres2amp tonek kres2amp, 10 
;	printf " Resonator 2 amp=%f\n", kres2amp, kres2amp

	kres3amp ctrl7 giChan, giRes3CC, 0.0, 1.0;, giExpr
	kres3amp tonek kres3amp, 10 
;	printf " Resonator 3 amp=%f\n", kres3amp, kres3amp

	kbrillamp ctrl7 giChan, giBrillCC, 0.0, 1.0;, giExpr
	kbrillamp tonek kbrillamp, 10 
;	printf " Brill amp=%f\n", kbrillamp, kbrillamp

	kfullamp ctrl7 giChan, giFullCC, 0.0, 1.0;, giExpr
	kfullamp tonek kfullamp, 10 
;	printf " Full amp=%f\n", kfullamp, kfullamp

	gkVamp ctrl7 giChan, giVibCC, 0.0, 10.0;, giExpr
;	printf " Vibrato level = %f\n", gkVamp, gkVamp

	kexpr ctrl7 giChan, giExprCC, 0.0, 1.0, giExpr
	kexpr tonek kexpr, 10 
;	printf " Expression = %f\n", kexpr, kexpr

	aDeep tone gaout, 400
	aRes1 reson gaout, 300, 200, 1
	aRes2 reson gaout, 800, 300, 1
	aRes3 reson gaout, 3000, 500, 1
	aBrill atone gaout, 2000
	aCombo mac kdeepamp,aDeep, kres1amp,aRes1, kres2amp,aRes2, kres3amp,aRes3, kbrillamp,aBrill, kfullamp, gaout
	gaout = aCombo*kexpr/6
endin


; Amplifier Overdrive Simulation:

alwayson "Overdrive"	; if you want (controllable) distortion

instr Overdrive
	pset 0, 0, 0, 0
	initc7 giChan, giOvrdrvCC, p4/4	; initc7 range is 0..1
	gkOverDrive init p4	; 1..4 = degree, 0 = OFF
	gkOverDrive ctrl7, giChan, giOvrdrvCC, 0, 4	; rescaled to max = 4
	; printk2 gkOverDrive

	; Overdrive distortion -- after Peter Billam's pjbsynth
	; -- in turn based on Steve Harris' LADSPA plugin 'foverdrive 1196':
	; drivem1 = drive - 1.0 ; fx = fabs(x);
	; output x*(fx + drive)/(x*x + drivem1*fx + 1.0);
	if gkOverDrive > 0.95 then
		kDm1 = gkOverDrive - 1
		aAbs = abs(gaout)
		gaout = gaout * ( aAbs + gkOverDrive ) \
			/ ( gaout*gaout + kDm1*aAbs + 1.0 )
	endif
endin


instr Output
	/** Stereo reverb:  Novachord doesn't have it, but can sound good...**/
	krevb ctrl7, 1, giRvbCC, 0, 0.8, giRvbRamp	; best >= 0.6, but can distort...
;	printf "krevb=%f\n", krevb, krevb
	
	if (krevb > 0) then
		arvbl, arvbr reverbsc gaout, gaout, krevb, 12000
		outs gaout+arvbl, gaout+arvbr
	else
		outs gaout, gaout
	endif
	gaout = 0
endin


instr Params
	fprints "nova_params",  "\n*** Controller Values: ***\n"

	ideepamp ctrl7 1, giDeepCC, 0, 127 
	fprints "nova_params", "Deep Tone\tCC %d = %d\n", giDeepCC, ideepamp

	if1amp ctrl7 1, giRes1CC, 0, 127
	fprints "nova_params", "Resonator 1\tCC %d = %d\n", giRes1CC, if1amp

	if2amp ctrl7 1, giRes2CC, 0, 127
	fprints "nova_params", "Resonator 2\tCC %d = %d\n", giRes2CC, if2amp

	if3amp ctrl7 1, giRes3CC, 0, 127
	fprints "nova_params", "Resonator 3\tCC %d = %d\n", giRes3CC, if3amp

	ibrillamp ctrl7 1, giBrillCC, 0, 127 
	fprints "nova_params", "Brilliant Tone\tCC %d = %d\n", giBrillCC, ibrillamp

	ifullamp ctrl7 1, giFullCC, 0, 127 
	fprints "nova_params", "Full Tone\tCC %d = %d\n", giFullCC, ifullamp

	fprints "nova_params", "\n"

	iatk ctrl7 1, giAtkCC, 0, 127
	fprints "nova_params", "Attack   \tCC %d = %d\n", giAtkCC, iatk

	isus ctrl7 1, giSusCC, 0, 127
	fprints "nova_params", "Sustain  \tCC %d = %d\n", giSusCC, isus

	irel ctrl7 1, giRelCC, 0, 127
	fprints "nova_params", "Release  \tCC %d = %d\n", giRelCC, irel

	idcy ctrl7 1, giDcyCC, 0, 127
	fprints "nova_params", "Decay    \tCC %d = %d\n", giDcyCC, idcy

	fprints "nova_params", "\n"

	ienvlp ctrl7 1, giEnvlpCC, 0, 127
	fprints "nova_params", "Attack Mode\tCC %d = %d\n", giEnvlpCC, ienvlp

	ishmode ctrl7 1, giHoldModeCC, 0, 127
	fprints "nova_params", "Hold Pedal Block\tCC %d = %d\n", giHoldModeCC, ishmode

	fprints "nova_params", "\n"

	ivamp ctrl7 1, giVibCC, 0, 127
	fprints "nova_params", "Vibrato  \tCC %d = %d\n", giVibCC, ivamp

	iexpr ctrl7 1, giExprCC, 0, 127
	fprints "nova_params", "Expression\tCC %d = %d\n", giExprCC, iexpr

	fprints "nova_params", "\n"

	irvrb ctrl7 1, giRvbCC, 0, 127
	fprints "nova_params", "Reverb  \tCC %d = %d\n", giRvbCC, irvrb

	iovdrv ctrl7 1, giOvrdrvCC, 0, 127
	fprints "nova_params", "Overdrive\tCC %d = %d\n", giOvrdrvCC, iovdrv

	fprints "nova_params", "\n"
endin

</CsInstruments>

<CsScore>
</CsScore>

</CsoundSynthesizer>
